import math
import cupy as cp
from cupyx import jit

@jit.rawkernel(device="cuda")
def get_pre_soma_spike(
    tick,
    agent_index,
    globals,
    agent_ids,
    pre_soma_id,
    t_current,
    input_spikes_tensor,  # input spikes
    output_spikes_tensor,
):
    t_current = int(tick)

    if not cp.isnan(pre_soma_id):
        i = 0
        while i < len(agent_ids) and agent_ids[i] != pre_soma_id:
            i += 1
        spike = output_spikes_tensor[i][t_current]
    else:
        spike = 0.0
        spike_buffer_max_len = len(input_spikes_tensor[agent_index])
        i = 0
        while i < spike_buffer_max_len and not cp.isnan(
            input_spikes_tensor[agent_index][i][0]
        ):
            if input_spikes_tensor[agent_index][i][0] == t_current:
                spike += input_spikes_tensor[agent_index][i][
                    1
                ]  # TODO: check if we need analog values for spikes
            i += 1
    
    # print(f"Pre-soma spike for agent {agent_index} at time {t_current}: {spike}")
    return spike


"""
Izhikevich Neuron and weighted synapse step functions for spiking neural networks

"""

from cupyx import jit


@jit.rawkernel(device="cuda")
def izh_soma_step_func(
    tick,
    agent_index,
    globals,
    agent_ids,
    breeds,
    locations,
    neuron_params,  # k, vth, C, a, b,
    internal_state,  # v, u
    synapse_history,  # Synapse delay
    input_spikes_tensor,  # input spikes
    output_spikes_tensor,
    internal_states_buffer,
):
    synapse_ids = locations[agent_index]  # network location is defined by neighbors

    I_synapse = 0.0
    for i in range(len(synapse_ids)):

        synapse_index = -1  # synapse index local to the current mpi rank
        i = 0
        while i < len(agent_ids) and agent_ids[i] != synapse_ids[0]:
            i += 1
        if i < len(agent_ids):
            synapse_index = i
            I_synapse += internal_state[synapse_index][0]

    # Get the current time step value:
    t_current = int(tick)

    dt = globals[0]  # time step size

    # NOTE: neuron_params would need to be as long as the max number of params in any spiking neuron model
    k = neuron_params[agent_index][0]
    vthr = neuron_params[agent_index][1]
    C = neuron_params[agent_index][2]
    a = neuron_params[agent_index][3]
    b = neuron_params[agent_index][4]
    vpeak = neuron_params[agent_index][5]
    vrest = neuron_params[agent_index][6]
    d = neuron_params[agent_index][7]
    vreset = neuron_params[agent_index][8]
    I_in = neuron_params[agent_index][9]

    # From https://www.izhikevich.org/publications/spikes.htm
    # v' = 0.04v^2 + 5v + 140 -u + I
    # u' = a(bv - u)
    # if v=30mV: v = c, u = u + d, spike

    # dv = (k*(internal_state[my_idx]-vrest)*(internal_state[my_idx]-vthr)-u[my_idx]+I) / C
    # internal_state: [0] - v, [1] - u
    # NOTE: size of internal_state would need to be set as the maximum possible state varaibles of any spiking neuron

    v = internal_state[agent_index][0]
    u = internal_state[agent_index][1]

    I_bias = globals[1]  # bias current

    dv = (k * (v - vrest) * (v - vthr) - u + I_synapse + I_bias + I_in) / C
    v = v + dt * dv *1e3

    u += dt * 1e3 * (a * (b * (v - vrest) - u))
    s = 1 * (v >= vthr)  # output spike
    u = u + d * s   # If spiked, update recovery variable
    v = v * (1 - s) + vreset * s  # If spiked, reset membrane potential

    internal_state[agent_index][0] = v
    internal_state[agent_index][1] = u

    output_spikes_tensor[agent_index][t_current] = s
    internal_states_buffer[agent_index][t_current][0] = v
    internal_states_buffer[agent_index][t_current][1] = u



"""
Model class for building an SNN
"""

from typing import Dict, Callable, List
from pathlib import Path

import numpy as np
from sagesim.space import NetworkSpace
from sagesim.model import Model
from sagesim.breed import Breed
from pathlib import Path

from superneuroabm.step_functions.soma.izh import izh_soma_step_func
from superneuroabm.step_functions.soma.lif import lif_soma_step_func
from superneuroabm.step_functions.synapse.single_exp import synapse_single_exp_step_func
from superneuroabm.step_functions.synapse.stdp.exp_pair_wise_stdp import (
    exp_stdp_all_to_all,
)

CURRENT_DIR_ABSPATH = Path(__file__).resolve().parent


class NeuromorphicModel(Model):
    def __init__(
        self,
        soma_breed_info: Dict[str, List[Callable]] = {
            "IZH_Soma": [
                (
                    izh_soma_step_func,
                    CURRENT_DIR_ABSPATH / "step_functions" / "soma" / "izh.py",
                )
            ],
            "LIF_Soma": [
                (
                    lif_soma_step_func,
                    CURRENT_DIR_ABSPATH / "step_functions" / "soma" / "lif.py",
                )
            ],
        },
        synapse_breed_info: Dict[str, List[Callable]] = {
            "Single_Exp_Synapse": [
                (
                    synapse_single_exp_step_func,
                    CURRENT_DIR_ABSPATH
                    / "step_functions"
                    / "synapse"
                    / "single_exp.py",
                )
            ],
        },
    ) -> None:
        """
        Creates an SNN Model and provides methods to create, simulate,
        and monitor soma and synapses.

        :param use_gpu: True if the system supports CUDA GPU
            acceleration.
        :param soma_breed_info: Dict of breed name to List of
            Callable step functions. If specifed, will override
            the default soma breed and soma and synapse step
            functions, allowing for multi-breed simulations.
            Step functions will be executed on the respective
            breed every simulation step in the order specifed in the
            list.
        """
        super().__init__(space=NetworkSpace())

        soma_properties = {
            "parameters": [0.0, 0.0, 0.0, 0.0, 0.0],  # k, vth, C, a, b,
            "internal_state": [0.0, 0.0, 0.0, 0.0],  # v, u
            "synapse_delay_reg": [],  # Synapse delay
            "input_spikes_tensor": [],  # input spikes tensor
            "output_spikes_tensor": [],
            "internal_states_buffer": [],
        }
        synapse_properties = {
            "parameters": [
                0.0 for _ in range(10)
            ],  # weight, delay, scale, Tau_fall, Tau_rise, tau_pre_stdp, tau_post_stdp, a_exp_pre, a_exp_post, stdp_history_length
            "internal_state": [
                0.0 for _ in range(4)
            ],  # Isyn, Isyn_supp, pre_trace, post_trace
            "synapse_delay_reg": [],  # Synapse delay
            "input_spikes_tensor": [],  # input spikes tensor
            "output_spikes_tensor": [],
            "internal_states_buffer": [],
        }
        self._synapse_ids = []
        self._soma_ids = []
        self._soma_reset_states = {}

        self._soma_breeds: Dict[str, Breed] = {}
        for breed_name, step_funcs in soma_breed_info.items():
            soma_breed = Breed(breed_name)  # Strt here Ashish
            for prop_name, default_val in soma_properties.items():
                soma_breed.register_property(prop_name, default_val)
            for step_func_order, (step_func, module_fpath) in enumerate(step_funcs):
                module_fpath = (
                    CURRENT_DIR_ABSPATH / "izh_soma.py"
                    if module_fpath is None
                    else module_fpath
                )
                soma_breed.register_step_func(
                    step_func=step_func,
                    module_fpath=module_fpath,
                    priority=step_func_order,
                )
            self.register_breed(soma_breed)
            self._soma_breeds[breed_name] = soma_breed

        self._synapse_breeds: Dict[str, Breed] = {}
        for breed_name, step_funcs in synapse_breed_info.items():
            synapse_breed = Breed(breed_name)  # Strt here Ashish
            for prop_name, default_val in synapse_properties.items():
                synapse_breed.register_property(prop_name, default_val)
            for step_func_order, (step_func, module_fpath) in enumerate(step_funcs):
                module_fpath = (
                    CURRENT_DIR_ABSPATH / "izh_soma.py"
                    if module_fpath is None
                    else module_fpath
                )
                synapse_breed.register_step_func(
                    step_func=step_func,
                    module_fpath=module_fpath,
                    priority=100 + step_func_order,
                )
            self.register_breed(synapse_breed)
            self._synapse_breeds[breed_name] = synapse_breed

        self._synapse_index_map = {}

    def set_global_property_value(name: str, value: float) -> None:
        if name in super().globals:
            super().set_global_property_value(name, value)
        else:
            super().register_global_property(name, value)

    def get_global_property_value(name: str) -> float:
        return super().get_global_property_value(name)

    def setup(
        self,
        use_gpu: bool = False,
        retain_parameters=False,
    ) -> None:
        """
        Resets the simulation and initializes agents.

        :param retain_parameters: False by default. If True, parameters are
            reset to their default values upon setup.
        """
        synapse_ids = self._synapse_ids
        soma_ids = self._soma_ids
        for synapse_id in synapse_ids:
            # Clear input spikes
            new_input_spikes = []
            super().set_agent_property_value(
                id=synapse_id,
                property_name="input_spikes_tensor",
                value=new_input_spikes,
            )
            # Clear synapse delay registers
            synapse_delay = len(
                super().get_agent_property_value(
                    id=synapse_id, property_name="synapse_delay_reg"
                )
            )
            synapse_delay_reg = [0 for _ in range(synapse_delay)]
            self.set_agent_property_value(
                synapse_id,
                "synapse_delay_reg",
                synapse_delay_reg,
            )
            # Reset parameters to defaults if retain_parameters is True
            if retain_parameters:
                # Reset all synapse parameters to their default values
                default_synapse_parameters = [
                    0.0 for _ in range(10)
                ]  # weight, delay, scale, Tau_fall, Tau_rise, tau_pre_stdp, tau_post_stdp, a_exp_pre, a_exp_post, stdp_history_length
                super().set_agent_property_value(
                    id=synapse_id,
                    property_name="parameters",
                    value=default_synapse_parameters,
                )
            # Clear internal states
            synapse_internal_state = super().get_agent_property_value(
                id=synapse_id,
                property_name="internal_state",
            )
            synapse_internal_state = [0.0 for _ in synapse_internal_state]
            super().set_agent_property_value(
                id=synapse_id,
                property_name="internal_state",
                value=synapse_internal_state,
            )
        for soma_id in soma_ids:
            # Clear internal states
            super().set_agent_property_value(
                id=soma_id,
                property_name="internal_state",
                value=self._soma_reset_states[soma_id],
            )
        super().setup(use_gpu=use_gpu)

    def simulate(
        self, ticks: int, update_data_ticks: int = 1  # , num_cpu_proc: int = 4
    ) -> None:
        """
        Override of superneuroabm.core.model mainly to register an
        AgentDataCollector to monitor marked output somas.

        """
        for soma_id in self._soma_ids:
            # Clear output buffer
            output_buffer = [0 for _ in range(ticks)]
            super().set_agent_property_value(
                id=soma_id,
                property_name="output_spikes_tensor",
                value=output_buffer,
            )
            initial_internal_state = super().get_agent_property_value(
                id=soma_id, property_name="internal_state"
            )
            internal_states_buffer = [initial_internal_state[::] for _ in range(ticks)]
            super().set_agent_property_value(
                id=soma_id,
                property_name="internal_states_buffer",
                value=internal_states_buffer,
            )
        for synapse_id in self._synapse_ids:
            initial_internal_state = super().get_agent_property_value(
                id=synapse_id, property_name="internal_state"
            )
            internal_states_buffer = [initial_internal_state[::] for _ in range(ticks)]
            super().set_agent_property_value(
                id=synapse_id,
                property_name="internal_states_buffer",
                value=internal_states_buffer,
            )
        super().simulate(ticks, update_data_ticks)  # , num_cpu_proc)

    def create_soma(
        self,
        breed: str,
        parameters: List[float],
        default_internal_state: List[float],
    ) -> int:
        """
        Creates and soma agent.

        :return: SAGESim agent id of soma

        """
        soma_id = super().create_agent_of_breed(
            breed=self._soma_breeds[breed],  # TODO fix
            parameters=parameters,
            internal_state=default_internal_state,
        )
        self._soma_ids.append(soma_id)
        self._soma_reset_states[soma_id] = default_internal_state
        return soma_id

    def create_synapse(
        self,
        breed: str,
        pre_soma_id: int,  # TODO edit
        post_soma_id: int,
        parameters: List[float],
        default_internal_state: List[float],
    ) -> int:
        """
        Creates and adds Synapse agent.

        :param pre_soma_id: int presynaptic soma id
        :param post_soma_id: int postsynaptic soma id
        :param weight: weight of synapse
        :param synaptic delay: number of timesteps to delay synapse by
        :param synapse_learning_params: Optional. Any parameters used in a learning
            enabled step function. Must be specified in order of use
            in step function.
        """
        synaptic_delay = int(parameters[1])
        delay_reg = [0 for _ in range(synaptic_delay)]
        synapse_id = self.create_agent_of_breed(
            breed=self._synapse_breeds[breed],
            parameters=parameters,
            internal_state=default_internal_state,
            synapse_delay_reg=delay_reg,
        )
        self._synapse_ids.append(synapse_id)

        network_space: NetworkSpace = self.get_space()
        if not np.isnan(pre_soma_id):
            network_space.connect_agents(synapse_id, pre_soma_id, directed=True)
        if not np.isnan(post_soma_id):
            network_space.connect_agents(post_soma_id, synapse_id, directed=True)
        return synapse_id

    def update_synapse(
        self,
        pre_soma_id: int,
        post_soma_id: int,
        parameters: List[float] = None,
    ):
        raise NotImplementedError("update_synapse is not implemented.")
        if pre_soma_id in self._synapse_index_map:
            output_synapse_index_map = self._synapse_index_map[pre_soma_id]
            if post_soma_id in output_synapse_index_map:
                synapse_idx = output_synapse_index_map[post_soma_id]
            else:
                raise ValueError(
                    f"Synapse {pre_soma_id} -> {post_soma_id} does not exist"
                )
        else:
            raise ValueError(f"Synapse {pre_soma_id} -> {post_soma_id} does not exist")

        # Update new synapse params
        if weight != None:
            output_synapses = self.get_agent_property_value(
                pre_soma_id, "output_synapses"
            )
            output_synapses[synapse_idx][1] = weight
            self.set_agent_property_value(
                pre_soma_id,
                "output_synapses",
                output_synapses,
            )

        # Update or enter learning params
        if synapse_learning_params:
            synapses_learning_params = self.get_agent_property_value(
                pre_soma_id, "output_synapses_learning_params"
            )
            synapses_learning_params[synapse_idx] = synapse_learning_params
            self.set_agent_property_value(
                id=pre_soma_id,
                property_name="output_synapses_learning_params",
                value=synapses_learning_params,
            )

    def add_spike(self, synapse_id: int, tick: int, value: float) -> None:
        """
        Schedules an external input spike to this soma.

        :param tick: tick at which spike should be triggered
        :param value: spike value
        """
        spikes = self.get_agent_property_value(
            id=synapse_id,
            property_name="input_spikes_tensor",
        )
        spikes.append([tick, value])
        self.set_agent_property_value(
            synapse_id, "input_spikes_tensor", spikes  # , [len(spikes), 2]
        )

    def get_spike_times(self, soma_id: int) -> np.array:
        spike_train = super().get_agent_property_value(
            id=soma_id,
            property_name="output_spikes_tensor",
        )
        spike_times = [i for i in range(len(spike_train)) if spike_train[i] > 0]
        return spike_times

    def get_internal_states_history(self, agent_id: int) -> np.array:
        return super().get_agent_property_value(
            id=agent_id, property_name="internal_states_buffer"
        )

    def summary(self) -> str:
        """
        Verbose summary of the network structure.

        :return: str information of netowkr struture
        """
        raise NotImplementedError("summary is not implemented.")
        summary = []
        summary.append("soma information:")
        for soma_id in range(self._agent_factory.num_agents):
            spikes = self.get_agent_property_value(soma_id, "output_spikes")
            summary.append(f"soma: {soma_id} Spike Train: {str(spikes)}")

        summary.append("\n\n\nSynapse information:")
        for presynaptic_soma_id in range(self._agent_factory.num_agents):
            synapses = self.get_agent_property_value(
                presynaptic_soma_id, "output_synapses"
            )
            for synapse in synapses:
                postsynaptic_soma_id = synapse[0]
                weight = synapse[1]
                summary.append(
                    (
                        f"soma {presynaptic_soma_id} -> {postsynaptic_soma_id}"
                        f": weight: {weight}"
                    )
                )

        return "\n".join(summary)

    def save(self, fpath: str):
        super().save(self, fpath)

    def load(self, fpath: str):
        self = super().load(fpath)



"""
Single exponential synapse step functions for spiking neural networks

"""

import numpy as np
from cupyx import jit

from superneuroabm.step_functions.synapse.util import get_pre_soma_spike


@jit.rawkernel(device="cuda")
def synapse_single_exp_step_func(
    tick,
    agent_index,
    globals,
    agent_ids,
    breeds,
    locations,
    synapse_params,  # scale, time constant (tau_rise and tau_fall)
    internal_state,  #
    synapse_history,  # delay
    input_spikes_tensor,  # input spikes
    output_spikes_tensor,
    internal_states_buffer,
):
    t_current = int(tick)

    dt = globals[0]  # time step size

    weight = synapse_params[agent_index][0]
    synaptic_delay = synapse_params[agent_index][1]
    scale = synapse_params[agent_index][2]
    tau_fall = synapse_params[agent_index][3]
    tau_rise = synapse_params[agent_index][4]

    location_data = locations[agent_index]
    if len(location_data) == 0:
        pre_soma_id = np.nan
    else:
        pre_soma_id = location_data[0]
    spike = get_pre_soma_spike(
        tick,
        agent_index,
        globals,
        agent_ids,
        pre_soma_id,
        t_current,
        input_spikes_tensor,
        output_spikes_tensor,
    )

    I_synapse = internal_state[agent_index][0]

    I_synapse = I_synapse * (1 - dt / tau_fall) + spike * scale * weight

    internal_state[agent_index][0] = I_synapse
    internal_states_buffer[agent_index][t_current][0] = I_synapse
    internal_states_buffer[agent_index][t_current][1] = spike
    internal_states_buffer[agent_index][t_current][2] = t_current



from setuptools import setup, find_packages

setup(
    name="SuperNeuroABM",
    version="1.0.0.dev1",
    author="Chathika Gunaratne, Shruti Kulkarni, Ashish Gautam, Xi Zhang, Prasanna Date",
    author_email="gunaratnecs@ornl.gov",
    packages=[
        "superneuroabm",
        "superneuroabm.step_functions",
        "superneuroabm.step_functions.soma",
        "superneuroabm.step_functions.synapse",
        "superneuroabm.step_functions.synapse.stdp",
    ],
    include_package_data=True,
    url="https://code.ornl.gov/superneuro/superneuroabm",
    license="GPL",
    description="A GPU-based multi-agent simulation framework for neuromorphic computing.",
    long_description="""A GPU-based multi-agent simulation framework for neuromorphic computing.""",
    long_description_content_type="text/markdown",
    project_urls={"Source": "https://github.com/ORNL/superneuroabm"},
    install_requires=[
        "sagesim",
    ],
)



"""
LIF Neuron and weighted synapse step functions for spiking neural networks

"""

from cupyx import jit


@jit.rawkernel(device="cuda")
def lif_soma_step_func(  # NOTE: update the name to soma_step_func from neuron_step_func
    tick,
    agent_index,
    globals,
    agent_ids,
    breeds,
    locations,
    neuron_params,  # k, vth, C, a, b,
    internal_state,  # v, u
    synapse_history,  # Synapse delay
    input_spikes_tensor,  # input spikes
    output_spikes_tensor,
    internal_states_buffer,
):
    synapse_ids = locations[agent_index]  # network location is defined by neighbors

    I_synapse = 0.0
    for i in range(len(synapse_ids)):

        synapse_index = -1  # synapse index local to the current mpi rank
        i = 0
        while i < len(agent_ids) and agent_ids[i] != synapse_ids[0]:
            i += 1
        if i < len(agent_ids):
            synapse_index = i
            I_synapse += internal_state[synapse_index][0]

    # Get the current time step value:
    t_current = int(tick)  # Check if tcount is needed or if we ca use this directly.
    dt = globals[0]  # time step size
    I_bias = globals[1]  # bias current

    # NOTE: neuron_params would need to as long as the max number of params in any spiking neuron model
    # Neuron Parameter
    C = neuron_params[agent_index][0]  # membrane capacitance
    R = neuron_params[agent_index][1]  # Leak resistance
    vthr = neuron_params[agent_index][2]  # spike threshold
    tref = neuron_params[agent_index][3]  # refractory period
    vrest = neuron_params[agent_index][4]  # resting potential
    vreset = neuron_params[agent_index][5]  # reset potential
    tref_allows_integration = neuron_params[agent_index][
        6
    ]  # whether to allow integration during refractory period
    I_in = neuron_params[agent_index][7]  # input current
    # vreset = neuron_params[agent_index][8]
    # I_in = neuron_params[agent_index][9]

    # NOTE: size of internal_state would need to be set as the maximum possible state varaibles of any spiking neuron
    # Internal state variables
    v = internal_state[agent_index][0]  # membrane potential
    tcount = internal_state[agent_index][1]  # time count from the start of the simulation
    tlast = internal_state[agent_index][2]  # last spike time

    # Calculate the membrane potential update
    dv = (vrest - v) / (R * C) + (I_synapse*1e-5 + I_bias + I_in) / C

    v +=  (dv*dt) if ((dt * tcount) > (tlast + tref)) or tref_allows_integration else 0.0

    s = 1 * (v >= vthr) and (
        dt * tcount > tlast + tref
    )  # output spike only happens if the membrane potential exceeds the threshold and the neuron is not in refractory period.
    tlast = tlast * (1 - s) + dt * tcount * s
    v = v * (1 - s) + vreset * s  # If spiked, reset membrane potential

    internal_state[agent_index][0] = v
    internal_state[agent_index][1] += 1
    internal_state[agent_index][2] = tlast
    internal_state[agent_index][3] = I_synapse  # Update time count

    output_spikes_tensor[agent_index][t_current] = s
    internal_states_buffer[agent_index][t_current][0] = internal_state[agent_index][0]
    internal_states_buffer[agent_index][t_current][1] = internal_state[agent_index][1]
    internal_states_buffer[agent_index][t_current][2] = internal_state[agent_index][2]
    internal_states_buffer[agent_index][t_current][3] = internal_state[agent_index][3]
